div.literal-block-wrapper {
    padding: 0;
}

span.caption-text {
    float: right;
}

span.red {
    color: red;
    font-weight: bold;
}

span.green {
    color: green;
    font-weight: bold;
}

img {
    max-width: none !important;
}

